class RemoteConfiguration(object):
    '''
    Configuration of an ssh connection with ssh_config(5)
    parameters made available
    '''
    def __init__(self, host, username=None, port=None):
        '''
        :param str host: Distant machine hostname
        :param str username: Distant machine login (if None, let ssh decide)
        :param int port: Connection port (if None, let ssh decide)
        '''
        self._host = host
        self._ssh_config = {}
        self.username = username
        self.port = port

    @property
    def username(self):
        ''' Username on distant host '''
        return self._ssh_config.get('User', None)

    def __set_property(self, key, value):
        if value is not None:
            self._ssh_config[key] = value
        else:
            self._ssh_config.pop(key, None)

    @username.setter
    def username(self, value):
        self.__set_property('User', value)

    @property
    def port(self):
        ''' Port to connect to '''
        return self._ssh_config.get('Port', None)

    @port.setter
    def port(self, value):
        self.__set_property('Port', value)

    def get(self, key):
        '''
        Get an ssh_config entry previously set
        :returns None if key is not set
        returns associated value if key is set
        '''
        if key in self._ssh_config:
            return self._ssh_config[key]
        return None

    def set(self, key, value):
        ''' Set an ssh_config entry '''
        self._ssh_config[key] = value

    def unset(self, key):
        ''' Remove an ssh_config entry '''
        self._ssh_config.pop(key, None)

    @property
    def host(self):
        ''' Host given to ssh command line '''
        return self._host

    def get_args(self):
        ''' Return all arguments as an array '''
        args = []
        for k, value in sorted(self._ssh_config.items()):
            # All ssh options take one argument
            args.append(u'-o')
            args.append(u'{k}={v}'.format(k=k, v=value))
        return args

    def __str__(self):
        return ' '.join(self.get_args())

    def __repr__(self):
        return str(self.get_args())



import logging

from .execution import Execution
from .reactor import Reactor

LOGGER = logging.getLogger(__name__)


class RemoteRun(Execution):
    '''
    This class is used to run a shell execution.
    '''

    def __init__(self, reactor, remote, cmd,
                 logfile=None):
        '''
        Create a new shell execution without starting it.

        :param Reactor reactor: The reactor used to execute monitors
        :param RemoteConfiguration remote: The remote configuration
        :param str cmd: str A binary or bash-compatible expression
        :param file logfile: A file object used to log shell execution output
        :param int remote_port: The ssh remote port number used
        '''
        super(RemoteRun, self).__init__(reactor, logfile)

        self.__remote = remote
        self.__cmd = cmd.replace(u'"', u'\\"')

        LOGGER.debug('created remote run "%s" with args "%s"',
                     self.__cmd,
                     remote)

    def start(self):
        '''
        Start the execution.
        '''
        self.__remote.set('BatchMode', 'yes')
        self.__remote.set('LogLevel', 'error')
        args = self.__remote.get_args()
        args.append(u'{host}'.format(host=self.__remote.host))
        args.append(u'bash -c "({0})< <(cat; pkill -P $$)"'.format(self.__cmd))
        self._start(u'ssh', args)

    def stop(self):
        '''
        Stop the execution.
        '''
        self._send_eof()


def remote_run(remote, cmd,
               logfile=None,
               ms_timeout=-1):
    '''
    Run `cmd` on remote.host as remote.username

    Log execution output into `logfile` if not None.
    Wait `ms_timeout` for execution to complete.

    :param RemoteConfiguration remote: The connection parameters
    :param str cmd: str A binary or bash-compatible expression
    :param file logfile: A file object used to log shell execution output
    :param int ms_timeout: Duration waited for an event to occur
    :rtype: int
    :return: The execution result code
    '''
    reactor = Reactor()
    proc = RemoteRun(reactor=reactor, remote=remote, cmd=cmd, logfile=logfile)

    with proc:
        while reactor.run(ms_timeout) > 0:
            pass

    return proc.result



import logging

from .execution import Execution
from .reactor import Reactor

LOGGER = logging.getLogger(__name__)


class RemoteCopy(Execution):
    '''
    This class is used to copy a file from localhost
    to a remote host.
    '''

    def __init__(self, reactor, remote,
                 local_path, remote_path,
                 logfile=None):
        '''
        Create a new shell execution without starting it.

        :param Reactor reactor: The reactor used to execute monitors
        :param RemoteConfiguration remote: The remote configuration
        :param str local_path: The file or directory local path
        :param str remote_path: The file or directory remote path
        :param file logfile: A file object used to log shell execution output
        '''
        super(RemoteCopy, self).__init__(reactor, logfile)

        self.__remote = remote
        self.__local_path = local_path.replace(u'"', u'\\"')
        self.__remote_path = remote_path.replace(u'"', u'\\"')

        LOGGER.debug('created copy "%s" -> "%s" on %s with options %s',
                     self.__local_path,
                     self.__remote_path,
                     self.__remote.host,
                     str(self.__remote))

    def start(self):
        '''
        Start the execution.
        '''
        self.__remote.set('BatchMode', 'yes')
        self.__remote.set('LogLevel', 'error')
        args = self.__remote.get_args()
        args += [u'-r',
                 u'{0}'.format(self.__local_path),
                 u'{0}:{1}'.format(self.__remote.host, self.__remote_path)]
        self._start(u'scp', args)

    def stop(self):
        '''
        Stop the execution.
        '''
        self._terminate()


def remote_copy(remote, local_path, remote_path,
                logfile=None,
                ms_timeout=-1):
    '''
    Copy `local_path` to `remote_path` on `remote.host`

    Log execution output into `logfile` if not None.
    Wait `ms_timeout` for execution to complete.

    :param remote a RemoteConfiguration object
    :param str local_path: The file or directory local path
    :param str remote_path: The file or directory remote path
    :param file logfile: A file object used to log shell execution output
    :param int ms_timeout: Duration waited for an event to occur
    :rtype: int
    :return: The execution result code
    '''
    reactor = Reactor()
    proc = RemoteCopy(reactor, remote, local_path, remote_path,
                      logfile=logfile)

    with proc:
        while reactor.run(ms_timeout) > 0:
            pass

    return proc.result



import logging

from .reactor import Reactor
from .remote_run import RemoteRun, remote_run
from .remote_copy import RemoteCopy, remote_copy
from .local_run import LocalRun, local_run
from .exceptions import AlreadyStartedException
from .remote_configuration import RemoteConfiguration

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = ['RemoteRun', 'remote_run', 'RemoteConfiguration',
           'RemoteCopy', 'remote_copy',
           'LocalRun', 'local_run',
           'Reactor',
           'AlreadyStartedException']

__version__ = '2.0.2'



import logging
import errno
import select
import weakref

LOGGER = logging.getLogger(__name__)


class Reactor(object):
    '''
    This class is used to execute execution(s) monitor(s).
    '''

    def __init__(self):
        '''
        Create a new reactor.
        '''
        self.poller = select.poll()
        self.fd_to_cmd = {}

    def register_execution(self, cmd):
        '''
        Register a new `cmd` on the reactor.

        This will allow reactor to monitor `cmd` output
        and execute `cmd` monitor accordingly.

        :param Execution cmd: The cmd to register
        '''
        self.poller.register(cmd.fileno(), select.POLLIN | select.POLLPRI)
        self.fd_to_cmd[cmd.fileno()] = weakref.ref(cmd)

        LOGGER.debug('registered %s', cmd)

    def unregister_execution(self, cmd):
        '''
        Unregister a `cmd`.

        :param Execution cmd: The cmd to unregister
        '''
        del self.fd_to_cmd[cmd.fileno()]
        self.poller.unregister(cmd)

        LOGGER.debug('unregistered %s', cmd)

    def _process_cmd_output(self, handle):
        weakcmd = self.fd_to_cmd.get(handle, None)

        cmd = weakcmd() if weakcmd else None

        if cmd:
            LOGGER.debug('%s has new output', cmd)
            cmd.process_output()

    def run(self, ms_timeout):
        '''
        Wait `ms_timeout` for some registered execution(s) to generate
        output and execute associated monitor(s).

        :param int ms_timeout: Duration waited for an event to occur
        :rtype: int
        :return: The event(s) count
        '''
        if not self.fd_to_cmd:
            return 0

        count = self.poller.poll(ms_timeout)
        for handle, __ in count:
            self._process_cmd_output(handle)

        return len(count) or -errno.ETIMEDOUT



import logging
import signal
import uuid
import pexpect

from .exceptions import AlreadyStartedException

LOGGER = logging.getLogger(__name__)


class Execution(object):
    '''
    This class is used to run a shell execution.
    '''

    def __init__(self, reactor, logfile):
        '''
        Create a new shell execution without starting it

        :param Reactor reactor: The reactor used to execute monitors
        :param file logfile: A file object used to log shell execution output
        '''
        self.__id = str(uuid.uuid4()).partition('-')[0]
        self.__reactor = reactor
        self.__exec = None

        self.__logfile = logfile

        self.__monitors = {}

        self.__result = None

        self.__register_new_line_monitor()
        self.__register_finalize()

        LOGGER.debug('created "%s"', self)

    def __del__(self):
        '''
        Stop the execution upon destruction.
        '''
        self.__finalize()

    def __enter__(self):
        '''
        Start the execution upon context enter.
        '''
        self.start()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        '''
        Stop the execution upon context exit.
        '''
        self.__finalize()
        return False

    def __register_new_line_monitor(self):
        def _on_new_line(run):
            pass

        self.register_monitor(u'\n', _on_new_line)

    def __register_finalize(self):
        def _on_exit(run):
            run.__finalize()

        self.register_exit_monitor(_on_exit)

    def __finalize(self):
        if not self.started():
            return

        self.__reactor.unregister_execution(self)
        self.__exec.close()

        if self.__exec.exitstatus is not None:
            self.__result = self.__exec.exitstatus
        else:
            self.__result = self.__exec.signalstatus

        self.__exec = None

        LOGGER.debug('finalized %s (%d)', self, self.__result)

    def _start(self, cmd, args=None):
        '''
        Start the execution.
        '''
        if self.started():
            raise AlreadyStartedException()

        self.__result = None
        self.__exec = pexpect.spawnu(cmd, args or [])

        self.__reactor.register_execution(self)

        LOGGER.debug('started %s', self)

    def _terminate(self):
        '''
        The execution is killed but any pending monitor(s)
        can still be called (e.g. on_exit).
        '''
        if not self.started():
            return

        self.__exec.kill(signal.SIGTERM)

        LOGGER.debug('terminated %s', self)

    def _send_eof(self):
        '''
        The execution is killed but any pending monitor(s)
        can still be called (e.g. on_exit).
        '''
        if not self.started():
            return

        self.__exec.sendeof()

        LOGGER.debug('terminated %s', self)

    def start(self):
        '''
        Start the execution.
        '''
        raise NotImplementedError

    @property
    def result(self):
        '''
        The return code of the execution.
        '''
        return self.__result

    def started(self):
        '''
        Check if the execution is started.

        :rtype: bool
        :return: True if started, False otherwise
        '''
        return self.__exec is not None

    def fileno(self):
        '''
        Return the execution output pipe fileno.

        :rtype: int
        :return: The pipe fileno
        '''
        return self.__exec.fileno()

    def register_monitor(self, pattern, callback):
        '''
        Register a `callback` to be executed once the `pattern` has matched
        execution output.

        :param str pattern: A pattern to match
        :param callable callback: A callback to invoke

        .. note:: The `pattern` can't match multi-lines execution output.
        '''
        self.__monitors.setdefault(pattern, []).append(callback)

        LOGGER.debug('registered monitor "%s" on %s',
                     self.__pattern_name(pattern), self)

    def register_exit_monitor(self, callback):
        '''
        Register `callback` to be executed once the execution has terminated.

        :param callable callback: A callback to invoke
        '''
        self.register_monitor(pexpect.EOF, callback)

    def process_output(self):
        '''
        Try to match execution output against registered monitor(s).
        '''
        patterns = [pexpect.TIMEOUT] + list(self.__monitors.keys())
        index = self.__exec.expect(patterns, timeout=0)

        if index:
            # ssh.before seems only valid when something
            # other that timeout matched.
            if self.__logfile:
                self.__logfile.write(self.__exec.before)

            self.__invoke_callbacks(patterns[index])

    def __str__(self):
        '''
        Return the string representation of the execution.

        :rtype: str
        :return: A string representation of the execution
        '''
        return 'execution <{0}>'.format(repr(self))

    def __repr__(self):
        '''
        Return the id of the execution.

        :rtype str:
        :return: An id
        '''
        return str(self.__id)

    def __invoke_callbacks(self, matched_pattern):
        LOGGER.debug('matched monitor "%s" on %s',
                     self.__pattern_name(matched_pattern),
                     self)

        for callback in self.__monitors[matched_pattern]:
            callback(self)

    @staticmethod
    def __pattern_name(monitor):
        if monitor is pexpect.EOF:
            return u'@eof@'
        return repr(monitor)



#!/usr/bin/env python

import os
import re
from setuptools import setup

HERE = os.path.dirname(__file__)


def _readme():
    with open(os.path.join(HERE, 'README.rst')) as f:
        return f.read()


def _version():
    with open(os.path.join(HERE, 'yassh', '__init__.py')) as f:
        return re.search(r"__version__ = '([^']+)'", f.read()).group(1)


setup(
    name='yassh',

    version=_version(),

    description='A remote process launcher.',
    long_description=_readme(),

    url='https://github.com/enyx-opensource/yassh.git',

    author='David Keller',
    author_email='david.keller@enyx.com',

    license='BSD',

    keywords='process launcher',

    install_requires=[
        'pexpect',
    ],

    tests_require=[
        'sure', 'behave', 'coverage',
    ],

    packages=['yassh'],
    platforms='UNIX'
)



import logging
from .execution import Execution
from .reactor import Reactor

LOGGER = logging.getLogger(__name__)


class LocalRun(Execution):
    '''
    This class is used to run locally a shell execution.
    '''

    def __init__(self, reactor, cmd, logfile=None):
        '''
        Create a new shell execution without starting it.

        :param Reactor reactor: The reactor used to execute monitors
        :param str cmd: A binary or bash-compatible expression
        :param file logfile: A file object used to log shell execution output
        '''
        super(LocalRun, self).__init__(reactor, logfile)

        self.__cmd = cmd.replace(u'"', u'\\"')

        LOGGER.debug('created local run "%s"', cmd)

    def start(self):
        '''
        Start the execution.
        '''
        args = ['-c', '({0})< <(cat; pkill -P $$)'.format(self.__cmd)]
        self._start(u'bash', args)

    def stop(self):
        '''
        Stop the execution.
        '''
        self._terminate()


def local_run(cmd, logfile=None, ms_timeout=-1):
    '''
    Run `cmd`.

    Log execution output into `logfile` if not None.
    Wait `ms_timeout` for execution to complete.

    :param str cmd: str A binary or bash-compatible expression
    :param file logfile: A file object used to log shell execution output
    :param int ms_timeout: Duration waited for an event to occur
    :rtype: int
    :return: The execution result code
    '''
    reactor = Reactor()
    proc = LocalRun(reactor, cmd, logfile)

    with proc:
        while reactor.run(ms_timeout) > 0:
            pass

    return proc.result
